/** @file freeGroups.cpp
 * Free groups with homo-, endo- and automorphisms.
 * 
 * @author Karsten Kremer
 * @version 1.0
 * @date 29.01.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */

#include <cstdio>
#include <cassert>
#include <sstream>
#include <list>
#include <algorithm>
#include "freeGroups.h"

using namespace std;

const string FreeGroup::generatorString(int x) const {
	string s;
	stringstream stream;
	stream << x;
	stream >> s;
	return s+" ";
}

int FreeGroup::countGenerator (int g) const {
	return count(begin(),end(),g) - count(begin(),end(),inverseGenerator(g));
}

bool FreeGroup::isFront(int x) const {
	if (isOne()) return false;
	return front() == x;
}

bool FreeGroup::isBack(int x) const {
	if (isOne()) return false;
	return back() == x;	
}
bool FreeGroup::isFrontInverse(int x) const {
	if (isOne()) return false;
	return front() == inverseGenerator(x);
}

bool FreeGroup::isBackInverse(int x) const {
	if (isOne()) return false;
	return back() == inverseGenerator(x);	
}

bool FreeGroup::firstMatches (const FreeGroup& other) const {
	if (other.isOne()) return false;
	else return isBackInverse(other.front());
}

bool FreeGroup::lastMatches (const FreeGroup& other) const {
	if (other.isOne()) return false;
	else return isFrontInverse(other.back());
}

void FreeGroup::push_back(int x) {
	assert (isGenerator(x));
	if (isBackInverse(x)) rep.pop_back();
	else rep.push_back(x);
}
void FreeGroup::push_front(int x) {
	assert (isGenerator(x));
	if (isFrontInverse(x)) rep.pop_front();
	else rep.push_front(x);
}

	
FreeGroup& FreeGroup::append(FreeGroup& other) {
	while (firstMatches(other)) {
		pop_back();
		other.pop_front();
	}
	rep.splice(rep.end(), other.rep);
	return (*this);
}

FreeGroup& FreeGroup::append_left(FreeGroup& other) {
	while (lastMatches(other)) {
		pop_front();
		other.pop_back();
	}
	rep.splice(rep.begin(), other.rep);
	return (*this);
}

FreeGroup& FreeGroup::mult(const FreeGroup& other) {
	FreeGroup buf = other;
	append(buf);
	return (*this);
}

FreeGroup& FreeGroup::mult_left(const FreeGroup& other) {
	FreeGroup buf = other;
	append_left(buf);
	return (*this);
}

FreeGroup& FreeGroup::invert() {
	list<int> result;
	for (iterator 	it = begin(); it != end(); it++)
		result.push_front(inverseGenerator(*it));
	rep = result;
	return (*this);
}

FreeGroup& FreeGroup::power(int n) {
	if (n < 0) {
		invert();
		n = -n;
	} 
	FreeGroup x = *this;
	setOne();
	for (int i = 1; i <= n; i++) mult(x);	
	return (*this);
}


void FreeGroup::reduce() {
	FreeGroup buf = *this;
	list<int>::iterator it;
	
	rep.clear();
	while(buf.getLength() > 0) {
		push_back(buf.front());
		buf.pop_front();
	}
}




ostream& operator << (ostream &os, const FreeGroup& f) {
	FreeGroup::iterator it;
	for (	it = f.begin(); it != f.end(); it++)
		os << f.generatorString(*it);
	return os;
}


// ******************************* F2

const string F2::generatorString(int x) const {
	assert (x >= 0 && x < 4);
	string result;
	if (x%2 == 0) result += xChar; else result += yChar;
	if (x >= 2) result += "'";
	return(result);
}

// ******************************* EndF2


bool EndF2::isInverse(const EndF2& other) const {
	return ((*this * other).isOne() && (other * *this).isOne());
}
		
EndF2& EndF2::mult(const EndF2& other) {
	F2 x = image(other.image(F2(0)));
	F2 y = image(other.image(F2(1)));
	set (x,y);
	return (*this);
}
EndF2& EndF2::mult_left(const EndF2& other) {
	F2 x = other.image(image(F2(0)));
	F2 y = other.image(image(F2(1)));
	set (x,y);
	return (*this);			
}

// ******************************* AutF2

void AutF2::set(const EndF2& f, const EndF2& finv) {
	assert(finv.isInverse(f));
	map = f; // Attention: this may change finv, if the pointers of map and finv coincide.
	inv = finv; // This has to be avoided.
}


AutF2& AutF2::mult(const AutF2& other) {
	map.mult(other.map);
	inv.mult_left(other.inv);
	return (*this);
}


AutF2& AutF2::power(int n) {
	if (n < 0) {
		invert();
		n = -n;
	} 
	AutF2 x = *this;
	setOne();
	for (int i = 1; i <= n; i++) mult(x);	
	return (*this);
}

