/** @file matrix.cpp
 * 2x2 integer matrices.
 * 
 * @author Karsten Kremer
 * @version 1.0
 * @date 29.01.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */
 
#include <iostream>
#include <cstdio>
#include <cassert>
#include "matrix.h"

using namespace std;

void IntMatrix2x2::set(int aa, int bb, int cc, int dd) {
	a = aa; b = bb; c = cc; d = dd;
}

void IntMatrix2x2::setMatrix(const IntMatrix2x2& other) {
	set(other.getA(),
		 other.getB(),
		 other.getC(),
		 other.getD());
}

bool IntMatrix2x2::equals (const IntMatrix2x2& other) const {
	return (getA()==other.getA() 
			&& getB()==other.getB() 
			&& getC()==other.getC() 
			&& getD()==other.getD());
}

IntMatrix2x2& IntMatrix2x2::scale(int x) {
	set(getA()*x, 
		 getB()*x, 
		 getC()*x, 
		 getD()*x);
	return (*this);
}

IntMatrix2x2& IntMatrix2x2::invert() {
	int D = det();
	assert (D==1 || D==-1);
	set(getD(), -getB(), -getC(), getA());
	scale(D);
	return (*this);
}

IntMatrix2x2& IntMatrix2x2::power(int n) {
	if (n < 0) {
		invert();
		n = -n;
	} 
	IntMatrix2x2 x = *this;
	setOne();
	for (int i = 1; i <= n; i++) mult(x);	
	return (*this);
}

IntMatrix2x2& IntMatrix2x2::add(const IntMatrix2x2& other) {
	set(getA()+other.getA(),
		 getB()+other.getB(),
		 getC()+other.getC(),
		 getD()+other.getD());
	return (*this);
}

IntMatrix2x2& IntMatrix2x2::sub(const IntMatrix2x2& other) {
	set(getA()-other.getA(),
		 getB()-other.getB(),
		 getC()-other.getC(),
		 getD()-other.getD());
	return (*this);	
}

IntMatrix2x2& IntMatrix2x2::mult(const IntMatrix2x2& other) {
	set(getA()*other.getA() + getB()*other.getC(),
		 getA()*other.getB() + getB()*other.getD(),
		 getC()*other.getA() + getD()*other.getC(),
		 getC()*other.getB() + getD()*other.getD());
	return (*this);
}

IntMatrix2x2& IntMatrix2x2::div(const IntMatrix2x2& other) {
	mult(other.inverse());
	return (*this);
}

ostream& operator << (ostream &os, const IntMatrix2x2& M) {
	os << "(" << M.getA() << "," << M.getB() << ";" << M.getC() << "," << M.getD() << ")";
	return os;
}

istream& operator >> (istream &is, IntMatrix2x2& M) {
	int a,b,c,d;
	is >> a >> b >> c >> d;
	M.set(a,b,c,d);
	return is;
}

