/** @file matrix.h
 * 2x2 integer matrices
 * 
 * @author Karsten Kremer
 * @version 1.2
 * @date 12.10.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Karsten Kremer
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: karsten.kremer@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */

#if !defined(ORIGAMI_MATRIX_H_INCLUDED)
#define ORIGAMI_MATRIX_H_INCLUDED

#include <cstdio>
#include <cassert>
#include <iostream>
#include <vector>

using namespace std;


class IntMatrix2x2 {
	friend class SL2_Z;
	protected:
		int a,b,c,d;
	public:
		IntMatrix2x2(int aa, int bb, int cc, int dd)	{set(aa,bb,cc,dd);}
		IntMatrix2x2(int x = 0)								{set(x,0,0,x);}
		IntMatrix2x2(const IntMatrix2x2& other) 		{setMatrix(other);}
		
		virtual ~IntMatrix2x2() 							{};		
		IntMatrix2x2 copy() const  						{return IntMatrix2x2(*this);}

		virtual void set(int aa, int bb, int cc, int dd);
		void setMatrix(const IntMatrix2x2& other); 
		void setOne() 											{set(1,0,0,1);}
		void setZero() 										{set(0,0,0,0);}
		
		bool equals(const IntMatrix2x2& other) const;
		bool isZero() const 									{return equals(IntMatrix2x2(0));}
		bool isOne() const 									{return equals(IntMatrix2x2(1));}
		bool isInvertible() const 							{int D = det(); return (D==1 || D==-1);}
		
		IntMatrix2x2& scale(int x);
		IntMatrix2x2& invert();
		IntMatrix2x2& power(int n);
		IntMatrix2x2& add(const IntMatrix2x2& other);
		IntMatrix2x2& sub(const IntMatrix2x2& other);
		IntMatrix2x2& mult(const IntMatrix2x2& other);
		IntMatrix2x2& div(const IntMatrix2x2& other);
				
		void operator = (const IntMatrix2x2& other)  {setMatrix(other);}
		void operator += (const IntMatrix2x2& other) {add(other);}
		void operator -= (const IntMatrix2x2& other) {sub(other);}
		void operator *= (const IntMatrix2x2& other) {mult(other);}
		void operator /= (const IntMatrix2x2& other) {div(other);}
		
		bool operator == (const IntMatrix2x2& other) const {return equals(other);}
		bool operator != (const IntMatrix2x2& other) const {return !equals(other);}

		IntMatrix2x2 operator + (const IntMatrix2x2& other) const {return copy().add(other);}
		IntMatrix2x2 operator - (const IntMatrix2x2& other) const {return copy().sub(other);}
		IntMatrix2x2 operator * (const IntMatrix2x2& other) const {return copy().mult(other);}
		IntMatrix2x2 operator / (const IntMatrix2x2& other) const {return copy().div(other);}
		IntMatrix2x2 operator ^ (int n) const 							{return copy().power(n);}
		IntMatrix2x2 inverse () const 										{return copy().invert();}
		IntMatrix2x2 operator - () const 									{return copy().scale(-1);}
		
		int getA () const {return a;}
		int getB () const {return b;}
		int getC () const {return c;}
		int getD () const {return d;}
		int det()	const {return (a*d - b*c);}
		int trace() const {return (a+d);}
		
};

ostream& operator << (ostream& os, const IntMatrix2x2& M);
istream& operator >> (istream& is, IntMatrix2x2& M); 


#endif // ORIGAMI_MATRIX_H_INCLUDED
