/** @file triangulierung.h
 * A triangulation of the origami-curve. Each triangle corresponds
 * to a copy of the standard fundamental domain of SL(2,Z).
 * The triangles are glued together by the algorithm for the 
 * computation of the Veech-Group in veechgruppe.h.
 * 
 * @author Gabriela Schmithuesen
 * @version 1.0
 * @date 29.01.2007
 */

/*  This file is part of the Origami-Library, which is a collection
 *  of algorithms that calculate properties of mathematical objects
 *  called "origamis".
 *
 *  Copyright (C) 2007 Gabriela Schmithuesen
 *  Address: Institute for Algebra and Geometry, Fakultaet fuer
 *  Mathematik, Universitaet Karlsruhe, D-76128 Karlsruhe, Germany.
 *  E-mail: schmithuesen@math.uni-karlsruhe.de
 *  or origami@math.uni-karlsruhe.de
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License
 *  as published by the Free Software Foundation; either version 3
 *  of the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this program; if not, see
 *  <http://www.gnu.org/licenses/> or write to the Free
 *  Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 *  Boston, MA  02110-1301, USA.
 *
 *  If you publish results you obtained by using this software please
 *  don't forget to give appropriate credit in your publication.
 */ 
 
#include "triangulierung.h"

// ***************+ dreieck

dreieck::dreieck(SL2_Z_ST vt) {
	dreieck_inhalt = vt; 
	Aerfasst = false;
	Berfasst = false; 
	Cerfasst = false;
   Awert = 0;
	Bwert = 0;
	Cwert = 0;
}
	
dreieck::dreieck() {
	dreieck_inhalt = SL2_Z_ST(F2()); 
	Aerfasst = false; 
	Berfasst = false; 
	Cerfasst = false;
	Awert = 0;
	Bwert = 0;
	Cwert = 0;
}


void dreieck::ausgabe_more1()  const {
		cout << "matrix:" << dreieck_inhalt << endl;
      cout << "Awert:" << Awert <<", "<< "Bwert:"<< Bwert<<", "<<"Cwert"<<Cwert<<"."<< endl;
		cout << "Aerfasst:"<<Aerfasst <<", "<< "Berfasst:"<< Berfasst<<", ";
		cout<< "Cerfasst"<<Cerfasst<<"."<< endl;
}


// ***************+ d_ecke


void d_ecke::ausgabe() const {
	list<dreieck*>::const_iterator it;
	cout << "Number of adjacent triangles: " << ordnung << endl;
	cout << "Number: " << nummer << endl;
	cout << "infinitely many? " << unendlich << endl;
	cout << "List of triangles: " << endl;
	for(it = dl.begin(); it != dl.end(); it++) (*it)->ausgabe();
	cout << endl;
}


// ***************+ triang

triang::triang(const dreieck& d) {
	fl.push_back(d);
   flaechenzahl = 1;
	eckenzahl = -1;
}

triang::triang() {
	fl.push_back(dreieck());
	flaechenzahl = 1;
	eckenzahl = -1;
}

void triang::ausgabe() const {
	cout << endl << endl;
   cout << "A system of coset representatives: " << endl;
	const_iterator it;
	
	for(it = begin();it != end();it++) {
		it->ausgabe_dreieck_inhalt();
		cout << endl;
	}
   
	cout << endl;
	cout << "Index of the Veech group: " << fl.size() << endl;
	cout << "Number of vertices: " << return_eckenzahl() << endl;
	cout << "Genus of the Teichmueller curve: " <<g() << endl;
	cout << "Number of Cusps: " << pkt_anzahl() << endl;
	cout << endl;
}


void triang::ausgabe_kurz()  const{
	cout << endl << endl;
   cout << "Number of triangles:" << fl.size() << endl;
	cout << "Number of vertices:" << return_eckenzahl() << endl;
	cout << "Genus:" <<g() << endl;
	cout << "Number of Punctures:" << pkt_anzahl() << endl;
	cout << endl;
}

  
void triang::ausgabe_ecken() const { //Neu!!! (28.8.2003)
	list<d_ecke>::const_iterator lauf;
	for (lauf = el.begin();lauf != el.end();lauf++) {
		cout << "Vertex:" << endl;
		lauf->ausgabe();
	}
}
 
 
void triang::ausgabe_spitzen() const { //Neu!!! (19.12.2003)
	list<d_ecke>::const_iterator lauf;
	dreieck* d;
	SL2_Z_ST A,P,P_lauf; 
	int breite,i,k;
	k = 0;
	for (lauf = el.begin();lauf != el.end();lauf++) { // run through all vertices
		if (lauf->return_unendlich()== true) {// print out only cusps
			k++; // Zaehler
			breite = lauf->return_ordnung();  
			d = lauf->dl.front();
			A = d->return_inhalt();
			// Calculation of the parabolic representative of the cusp:
			P = A*SL2_Z_ST(F2(1))*A.inverse(); P_lauf.setOne();
			for (i=1; i <= breite ; i++)
				P_lauf = P*P_lauf; 

			//Output:
	  		cout << endl << k << "-th cusp: " << endl; 
	  		cout << "Amplitude: " << breite << endl;
	  		cout << "Parabolic representative: " << P_lauf;
	  		cout << endl; 
		}
	}
}
  

void triang::ausgabe_flaechen_more1() const { //Neu!!! (1.9.2003)
	const_iterator lauf;
	for (lauf = begin();lauf != end();lauf++) {
	  cout << "Surface:" << endl;
	  lauf->ausgabe_more1();
	}
}

dreieck* triang::anfuegen(int translattyp, const dreieck& d_neu, dreieck* dd_alt) {
	flaechenzahl++ ;
	fl.push_back(d_neu);
	dreieck* dd_neu = &fl.back();
	if (translattyp == 0) {
		dd_alt->def_stranslat(dd_neu);
		dd_neu->def_stranslat(dd_alt);
	};
	if (translattyp == 1) {
		dd_alt->def_ttranslat(dd_neu);
		dd_neu->def_tinvtranslat(dd_alt);
	};
	if (translattyp == 3) {
		dd_alt->def_tinvtranslat(dd_neu);
		dd_neu->def_ttranslat(dd_alt);
	};
	return dd_neu;
}

   //"verkleben" glues the triangles d_ausgang and d_ziel along the edge in direction
   //translattyp (similarly as in "anfuegen") of the triangle d_ausgang
void triang::verkleben (int translattyp, dreieck* dd_ausgang, dreieck* dd_ziel) {
	if (translattyp == 0) {
		dd_ausgang->def_stranslat(dd_ziel);
		dd_ziel->def_stranslat(dd_ausgang);
	};
	if (translattyp == 1) {
		dd_ausgang->def_ttranslat(dd_ziel);
		dd_ziel->def_tinvtranslat(dd_ausgang);
	};
	if (translattyp == 3) {
		dd_ausgang->def_tinvtranslat(dd_ziel);
		dd_ziel->def_ttranslat(dd_ausgang);
	};
};

   //"ecken_bestimmen" determines the list of vertices ("Eckenliste") 
   //of the triangulated surface in the following way: 
   //Iteratively for each vertex of a triangle which is not 
   //marked (erfasst = false), it determines all vertices,which are glued to this vertex.

void triang::ecken_bestimmen () {
      //Variables:
	dreieck d, d_rennt;
	d_ecke A,B,C;
	iterator dd;
	dreieck *dd_rennt; 
	int k, marker;
	bool fertig;

	k = 0; eckenzahl = 0; sondereckenzahl = 0;

     //Assign to all vertices of triangles that they are unmarked (erfasst = false)
	for (dd = fl.begin(); dd != fl.end(); dd++) {
		dd->def_Aerfasst(false);
		dd->def_Berfasst(false);
		dd->def_Cerfasst(false);
		dd->def_Awert(0);
		dd->def_Bwert(0);
		dd->def_Cwert(0);
	};

     //Find for all marked vertices, all vertices which are glued with them.
	for (dd = fl.begin(); dd != fl.end(); dd++)	{

        //Test vertex A ("Ecke A"):
		if (dd->return_Aerfasst() == false) {
			k++;
          //d_rennt = d;
			dd_rennt = &*dd;
			A = d_ecke(k,true); 
			el.push_back(A);

			do {

		     //In the triangle d_rennt mark its vertex A:
				dd_rennt->def_Awert(k);
				dd_rennt->def_Aecke(&(el.back()));
				dd_rennt->def_Aerfasst(true);
	     //Put the triangle d into this vertex:
             //(In der Ecke wird das Dreieck d eingetragen:)
				(el.back()).einfuegen(&(*dd_rennt));
             //d continues running in the direction T:
				dd_rennt = (dd_rennt->return_ttranslat());

	
				if (dd_rennt->return_Aerfasst()  && (dd_rennt->return_Awert() != k))	{
					cout << "Error in Case A:" << endl <<
							"(*dd_rennt).return_Awert() = "  << (*dd_rennt).return_Awert() <<" , " <<
							"k = " << k << endl ;
					return;
				}
			}	while(dd_rennt->return_Aerfasst() == false);
		}



         //Test of vertex B ("Ecke B"):
		fertig = false; //terminating condition ("Abbruchbedingung")
		if (dd->return_Berfasst() == false) {
			k++;
          //d_rennt = d;
			dd_rennt = &*dd;
         marker = 0; //Mark which kind of change-over ("Markierung welche Art von Ubergang"):
                      //0 = Tinv, 1 = S
			B = d_ecke(k, false); el.push_back(B);

			do	 {
				if (marker == 0) {

               //In the triangle d_rennt the vertex is put into B:
					dd_rennt->def_Bwert(k);
					dd_rennt->def_Becke(&(el.back()));
					dd_rennt->def_Berfasst(true);
               //The triangle d is put into this vertex:
					(el.back()).einfuegen(dd_rennt);
               //d continues running in the direction Tinv:
					dd_rennt = dd_rennt->return_tinvtranslat();
					//(*dd_rennt).ausgabe_more1();

					if ((*dd_rennt).return_Cerfasst()  && ((*dd_rennt).return_Cwert() != k)) {
						cout << "Error in Case B_1:" << endl <<
					            "(*dd_rennt).return_Cwert() = "  << (*dd_rennt).return_Cwert() <<" , " <<
								"k = " << k << endl;
						return;
					}
					if ((*dd_rennt).return_Cerfasst() == true) fertig = true;
			 	}
			 	if (marker == 1) {
               //This vertex is put in at C in the triangle d_rennt:
					dd_rennt->def_Cwert(k);
					dd_rennt->def_Cecke(&(el.back()));
					dd_rennt->def_Cerfasst(true);
               //The triangle d is put into this vertex:
					(el.back()).einfuegen(dd_rennt);
               //Is there a special vertex (i.e. a fixed point of a conjugate of S)?
					if (dd_rennt == (*dd_rennt).return_stranslat()) sondereckenzahl++;
				 	dd_rennt = dd_rennt->return_stranslat();

				 	if (dd_rennt->return_Berfasst()  && (dd_rennt->return_Bwert() != k)) {
						cout << "Error in Case B_2:" << endl <<
							 "(*dd_rennt).return_Bwert() = "  << dd_rennt->return_Bwert() <<" , " <<
							 "k = " << k << endl;
					 	return;
					}
					if ((*dd_rennt).return_Berfasst() == true) fertig = true;
			 	}
			 	
				marker = 1 - marker;
			} while(fertig == false);
		}

      //Test of the vertex C ("Ecke C"):
		fertig = false; //terminating condition ("Abbruchbedingung")
		if (dd->return_Cerfasst() == false) {
			k++;
          //d_rennt = d;
			dd_rennt = &*dd;
         marker = 1; //Mark which kind of change-over ("Markierung welche Art von Uebergang"):
                      //0 = Tinv, 1 = S
			B = d_ecke(k, false); 
			el.push_back(B);

			do {
				if (marker == 0) {
               //This vertex is put in at B in the triangle d_rennt:
					dd_rennt->def_Bwert(k);
				 	dd_rennt->def_Becke(&(el.back()));
				 	dd_rennt->def_Berfasst(true);
               //The triangle d is put into this vertex:
				 	(el.back()).einfuegen(dd_rennt);
               //d continues running in the direction Tinv:
				 	dd_rennt = dd_rennt->return_tinvtranslat();
				 	if (dd_rennt->return_Cerfasst()  && (dd_rennt->return_Cwert() != k)) {
						cout << "Error in Case C_1:" << endl <<
							 "(*dd_rennt).return_Cwert() = "  << (*dd_rennt).return_Cwert() <<" , " <<
							 "k = " << k << endl;
					 	return;
					}
					if (dd_rennt->return_Cerfasst() == true) fertig = true;
			 	}
			 	if (marker == 1) {
               //This vertex is put in at C in the triangle d_rennt:
				 	dd_rennt->def_Cwert(k);
				 	dd_rennt->def_Cecke(&(el.back()));
				 	dd_rennt->def_Cerfasst(true);
               //The triangle d is put into the vertex:
				 	(el.back()).einfuegen(dd_rennt);
               //Is there a special vertex (i.e. a fixed point of a conjugate of S)?
				 	if (dd_rennt == (*dd_rennt).return_stranslat()) sondereckenzahl++;
               //d continues running in the direction S:
				 	dd_rennt = (*dd_rennt).return_stranslat();
				 	if (dd_rennt->return_Berfasst()  && (dd_rennt->return_Bwert() != k))	 {
						cout << "Error in Case C_2:" << endl <<
							 "(*dd_rennt).return_Bwert() = "  << (*dd_rennt).return_Bwert() <<" , " <<
							 "k = " << k << endl;
					 	return;
					}
					if (dd_rennt->return_Berfasst() == true) fertig = true;
			 	}
			 	
				marker = 1 - marker;
			} while(fertig == false);
		}

	}
	eckenzahl = el.size();
}


int triang::g() const {// Variables:
	int chi, ecken, kanten, flaechen, geschlecht;

	ecken = eckenzahl + sondereckenzahl;
	flaechen = flaechenzahl + sondereckenzahl;
	kanten = 3*flaechen/2;
	chi = ecken - kanten + flaechen;
	geschlecht = (2 - chi)/2;
	return geschlecht;
}


//pkt_anzahl returns the number of punctures (i.e. vertices at infinity)
int triang::pkt_anzahl() const {
	list<d_ecke>::const_iterator ee;
   int pkt;

   //Run through the list of vertices ("Eckenliste") 
   //and increase by 1, if the vertex is at infinity:
	pkt = 0;
	for (ee = el.begin(); ee != el.end(); ee++)
		if (ee->return_unendlich()) pkt++;

	return pkt;
};



